/*===========================================================================
 * demo2019.c version 1.0
 * Copyright (C)2000 Micro Technic A-S
 *
 * This is a small demonstration program that shows the usage of the
 * functions in pv2019.c.
 * For information about the PV-2019, see http://www.micro-technic.com/pv2019
 *
 * It requires the base address of the PV-2019 as command line parameter, as
 * set on SW1 on the adapter.
 * Apart from calling all the functions, the program does nothing
 * particularly useful.
 */

#include <stdio.h>
#include <stdlib.h>
#include "pv2019.h"


int base;


void syntax (void)
{
  printf ("Syntax: DEMO2019 base\n");
  printf ("where base is the base address of the PV2019 adapter.\n");
  printf ("base can be one of these:\n");
  printf ("  200 208 210 218 220 228 230 238 300 308 310 318 320 328 330 338\n");
  exit (1);
}



int main (int argc, char **argv)
{
  char *endptr;
  unsigned int value;
  unsigned char cntstatus;
  int i;

  if (argc != 2)
    syntax();

  /* Read the base addr from command line parameter */
  base = (int) strtol (argv[1], &endptr, 16);
  if (*endptr != '\0')
    syntax();

  /* Check that a PV2019 really is at this base address. Under Linux, you
   * must call this function before calling other PV2019 functions. Under
   * other systems, you don't have to call it, but you may wish to do it
   * in order to verify the presence of a PV2019 adapter. */
  value = PV2019_Init (base);
  switch (value)
  {
    case PV2019_ERR_ILLEGAL_BASE:
           printf ("Error: Illegal base address\n");
           return 1;

    case PV2019_ERR_NOT_ROOT:
           /* Linux specific error */
           printf ("Error: Couldn't get I/O privileges. This program must run as root.\n");
           return 1;

    case PV2019_ERR_NOT_FOUND:
           printf ("Error: No PV2019 adapter found at this base address\n");
           return 1;

    case PV2019_ERR_MPU:
           printf ("Error: No working PV2019 adapter found at this base address\n");
           return 1;

    default:
           printf ("PV2019 adapter found at %X, MPU firmware version %u.%02u\n",
                   base, value/256, value%256);
           break;
  }

  /* Show the eight individual digital input bits */
  printf ("DI 7-0 =");
  for (i=7; i>=0; i--)
    printf (" %d", PV2019_GetInputBit (base, i));
  printf ("\n");

  /* Show the eight digital input bits as one hex byte */
  printf ("DI = 0x%02X\n", PV2019_GetInputByte (base));

  /* Turn digital output DO0 on, DO4 off, and DO7 on */
  PV2019_SetOutputBit (base, 0, 1);
  PV2019_SetOutputBit (base, 4, 0);
  PV2019_SetOutputBit (base, 7, 1);

  /* Show the last value written to digital output DO7 */
  printf ("DO7 = %d\n", PV2019_GetOutputBit (base, 7));

  /* Set all eight digital outputs DO0-7 at once: DO0 and DO4 on, others off */
  PV2019_SetOutputByte (base, 0x11);

  /* Show the last value written to the digital outputs as one hex byte */
  printf ("DO0-7 = 0x%02X\n", PV2019_GetOutputByte (base));

  /* Show voltage on AIN4, using the 10V range and 16 averages */
  value = PV2019_AIN_Volt (base, 4, AIN_10V, AVG_16);
  printf ("AIN4  = %4u = %6.3lf V\n", value, value*10.0/4095);

  /* Show voltage on AIN11, using the 2.5V range and 1 average */
  value = PV2019_AIN_Volt (base, 11, AIN_2V5, AVG_1);
  printf ("AIN11 = %4u = %6.3lf V\n", value, value*2.5/4095);

  /* Show current on AIN16, using 8 averages */
  value = PV2019_AIN_Curr (base, 16, AVG_8);
  printf ("AIN16 = %4u = %6.3lf mA\n", value, value*20.0/4095);

  /* Set AOUT0 to 7.5V (three quarters of the 0-10V range) */
  PV2019_AOUT (base, 0, AOUT_0_10V, 3072);

  /* Set AOUT0 to 5 mA (one quarter in the 0-20mA range) */
  PV2019_AOUT (base, 0, AOUT_0_20MA, 1024);

  /* Set AOUT1 to 12 mA (halfways in the 4-20mA range) */
  PV2019_AOUT (base, 1, AOUT_4_20MA, 2048);

  /* Show the actual value of the two counter inputs */
  cntstatus = PV2019_CntStatus (base);
  if (cntstatus & 0x01)		/* Check bit 0 */
    printf ("Counter 0 input is high\n");
  else
    printf ("Counter 0 input is low\n");

  if (cntstatus & 0x02)		/* Check bit 1 */
    printf ("Counter 1 input is high\n");
  else
    printf ("Counter 1 input is low\n");


  /* Reset and start counter 0 */
  PV2019_CntReset (base, 0);
  PV2019_CntEnable (base, 0);

  /* Stop counter 0 and read its value */
  printf ("Counter 0 = %lu\n", PV2019_CntDisable (base, 0));

  /* Reset and start counter 1 */
  PV2019_CntReset (base, 1);
  PV2019_CntEnable (base, 1);

  /* Read counter 1 value, and let it continue counting */
  printf ("Counter 1 = %lu\n", PV2019_CntEnable (base, 1));

  return 0;
}
