/* ==========================================================================
 * pv2019.c version 1.0
 * Copyright (C)2000 Micro Technic A-S
 *
 *       C/C++ interface for the Micronix PV-2019 adapter
 *
 * This file is a simple, thin interface to the analog I/Os, digital I/Os,
 * and the counters of the PV-2019 adapter.  You can use it as is, or as a
 * starting point for your own code.
 * For information about the PV-2019 adapter and the I/O registers it
 * implements, see http://www.micro-technic.com/pv2019
 *
 * The functions can be used with multiple PV-2019 adapters simultaneously
 * as all functions take the desired adapter's base address as a parameter.
 *
 * Reentrancy: With multitasking operating systems, you can use different
 * PV-2019 adapters from different threads/processes simultaneously.  You
 * should not communicate with the same PV-2019 adapter from different
 * threads/processes simultaneously, unless you protect the PV-2019
 * communication by mutex semaphores or other mechanisms.
 *
 * This file can be used with various C and C++ compilers, and various
 * operating systems that allow direct I/O operations.  It cannot be used
 * from Windows NT applications since Windows NT doesn't allow direct I/O.
 *
 * Usage: See the comments at each function, and demo2019.c.
 *
 * Compilers:
 *   Borland C/C++, Turbo C/C++
 *   Microsoft C/C++, Visual C++
 *   Watcom C/C++
 *   GNU GCC under Linux
 *
 * Operating systems:
 *   DOS (16-bit)
 *   DOS with DOS extender
 *   Windows 3.x
 *   Windows 9x
 *   Linux (application must run with root privileges)
 */

#include "pv2019.h"


/* Direct I/O is done differently in different compilers */

#ifdef __GNUC__
  /* Direct port I/O in Linux/GCC */
  #define COMPILER_RECOGNIZED
  #include <unistd.h>
  #include <asm/io.h>
  #define IN8(port)         inb((unsigned short)(port))
  #define IN16(port)        inw((unsigned short)(port))
  #define OUT8(port,value)  outb((value),(unsigned short)(port))
  #define OUT16(port,value) outw((value),(unsigned short)(port))
#endif

#ifdef _MSC_VER
  /* Direct port I/O in Microsoft C/C++, Visual C++ */
  #define COMPILER_RECOGNIZED
  #include <conio.h>
  #define IN8(port)         _inp((unsigned short)(port))
  #define IN16(port)        _inpw((unsigned short)(port))
  #define OUT8(port,value)  _outp((unsigned short)(port),(value))
  #define OUT16(port,value) _outpw((unsigned short)(port),(value))
#endif

#ifdef __TURBOC__
  /* Direct port I/O in Borland C/C++, Turbo C/C++ */
  #define COMPILER_RECOGNIZED
  #include <dos.h>
  #define IN8(port)         inportb((unsigned short)(port))
  #define IN16(port)        inport((unsigned short)(port))
  #define OUT8(port,value)  outportb((unsigned short)(port),(value))
  #define OUT16(port,value) outport((unsigned short)(port),(value))
#endif

#ifdef __WATCOMC__
  /* Direct port I/O in Watcom C 16/32 bit */
  #define COMPILER_RECOGNIZED
  #include <conio.h>
  #define IN8(port)         inp((unsigned)(port))
  #define IN16(port)        inpw((unsigned)(port))
  #define OUT8(port,value)  outp((unsigned)(port),(value))
  #define OUT16(port,value) outpw((unsigned)(port),(value))
#endif

#ifndef COMPILER_RECOGNIZED
  #error Don't know how to do direct I/O with this compiler!
#endif


/* I/O port addresses */
#define REG_SIGNATURE	0	/* read only  */
#define REG_DIGOUT	1	/* read/write */
#define REG_DIGIN	2	/* read only  */
#define REG_MPU_STATUS	3	/* read only  */
#define REG_CMDRSP_LO	4	/* read/write */
#define REG_CMDRSP_HI	5	/* read/write */

/* Expected value in REG_SIGNATURE register */
#define SIGNATURE       0x85

/* Bits in REG_MPU_STATUS */
#define BUSY            0x01



/*---------------------------------------------------------------------------
 * PV2019_Cmd(): This function sends a 16-bit command to the MPU and receives
 * the 16-bit result.  In case of timeout, the function returns 0.
 * The function is mostly used in this module, but it can also be called
 * directly from applications.
 */
unsigned short PV2019_Cmd (int base, unsigned short cmd)
{
  int timeout;

  OUT16 (base+REG_CMDRSP_LO, cmd);

  /* Each time around in the poll loop, we read the MPU status register.
   * According to IEEE P996, an 8-bit ISA I/O read cycle takes at least
   * 689 ns, so even with an infinitely fast CPU, this timeout will be long
   * enough. */
  timeout = 500;
  while (IN8 (base+REG_MPU_STATUS) & BUSY)
  {
    /* MPU still busy processing the command */
    if (--timeout == 0)
      return 0u;
  }
  return IN16 (base+REG_CMDRSP_LO);
}



/*---------------------------------------------------------------------------
 * PV2019_Init(): This function verifies the presence of a PV-2019 adapter
 * at the given base address.  For Linux, this function must be called
 * before using the other functions.  For other OSes, it isn't strictly
 * necessary to to call this function, but your application can give a
 * meaningful error message if the PV-2019 isn't present.
 *
 * Parameters:
 *   base = base address of the PV-2019 adapter as set on SW1
 *
 * Return values:
 *   PV2019_ERR_ILLEGAL_BASE: illegal value for base
 *   PV2019_ERR_NOT_ROOT    : Linux specific: ioperm() call failed
 *   PV2019_ERR_NOT_FOUND   : No PV-2019 found at this base address
 *   PV2019_ERR_MPU         : Timeout communicating with the MPU
 *   Other values           : OK, MPU firmware version number
 */
unsigned short PV2019_Init (int base)
{
  /* Check that base is one of the legal ones:
   *   0x200, 0x208, 0x210, 0x218, 0x220, 0x228, 0x230, 0x238,
   *   0x300, 0x308, 0x310, 0x318, 0x320, 0x328, 0x330, 0x338 */
  if ((base & 0xFEC7u) != 0x0200u)
    return PV2019_ERR_ILLEGAL_BASE;

#ifdef __GNUC__
  /* Under Linux, enable direct I/O access in the PV2019 I/O range.
   * This requires root privileges. */
  if (ioperm (base, 8, 1))
    return PV2019_ERR_NOT_ROOT;	/* Probably not running as root */
#endif

  /* Check that a PV2019 is actually present at this address */
  if (IN8 (base+REG_SIGNATURE) != SIGNATURE)
    return PV2019_ERR_NOT_FOUND;

  /* Get the MPU firmware version number. If this communication fails, e.g.
   * if it really isn't a PV2019, or the MPU isn't working, the return
   * value will be 0. */
  return PV2019_Cmd (base, 0x4F00u);
}



/*---------------------------------------------------------------------------
 * PV2019_GetInputBit(): Get one digital input bit
 *
 * Parameters:
 *   base  = base address of the PV-2019 adapter
 *   bitno = Bit number (0-7)
 *
 * Return value: the digital input bit (0 or 1)
 */
int PV2019_GetInputBit (int base, int bitno)
{
  bitno &= 0x0007;
  return (IN8 (base+REG_DIGIN) >> bitno) & 0x01;
}



/*---------------------------------------------------------------------------
 * PV2019_GetInputByte(): Get all eight digital input bits
 *
 * Parameters:
 *   base = base address of the PV-2019 adapter
 *
 * Return value: the eight input bits (0x00 - 0xFF)
 */
unsigned char PV2019_GetInputByte (int base)
{
  return IN8 (base+REG_DIGIN);
}



/*---------------------------------------------------------------------------
 * PV2019_SetOutputBit(): Set/reset one digital output bit
 *
 * Parameters:
 *   base  = base address of the PV-2019 adapter
 *   bitno = number of the digital output to set/reset (0-7)
 *   value = set the bit (1) or reset it (0)
 */
void PV2019_SetOutputBit (int base, int bitno, int value)
{
  bitno &= 0x0007;
  if (value)
  {
    /* Set the output bit */
    OUT8 (base+REG_DIGOUT, IN8 (base+REG_DIGOUT) | (1 << bitno));
  }
  else
  {
    /* Clear the output bit */
    OUT8 (base+REG_DIGOUT, IN8 (base+REG_DIGOUT) & ~(1 << bitno));
  }
}



/*---------------------------------------------------------------------------
 * PV2019_GetOutputBit(): Read the current value of one digital output bit,
 * as last set by PV2019_SetOutputBit() or PV2019_SetOutputByte().
 *
 * Parameters:
 *   base  = base address of the PV-2019 adapter
 *   bitno = number of the digital output to read (0-7)
 *
 * Return value: the digital output bit (0 or 1)
 */
int PV2019_GetOutputBit (int base, int bitno)
{
  bitno &= 0x0007;
  return (IN8 (base+REG_DIGOUT) >> bitno) & 0x01;
}



/*---------------------------------------------------------------------------
 * PV2019_SetOutputByte(): Set all eight digital output bits simultaneously.
 *
 * Parameters:
 *   base  = base address of the PV-2019 adapter
 *   value = the eight output bits (0x00 - 0xFF)
 */
void PV2019_SetOutputByte (int base, unsigned char value)
{
  OUT8 (base+REG_DIGOUT, value);
}



/*---------------------------------------------------------------------------
 * PV2019_GetOutputByte(): Read the current value of all eight digital output
 * bits, as last set by PV2019_SetOutputBit() or PV2019_SetOutputByte().
 *
 * Parameters:
 *   base  = base address of the PV-2019 adapter
 *
 * Return value: the eight output bits (0x00 - 0xFF)
 */
unsigned char PV2019_GetOutputByte (int base)
{
  return IN8 (base+REG_DIGOUT);
}



/*---------------------------------------------------------------------------
 * PV2019_AIN_Volt(): Read one of the analog voltage inputs (AIN 0-15).
 *
 * Parameters:
 *   base    = base address of the PV-2019 adapter
 *   channel = AIN input number (0-15)
 *   range   = Voltage range (AIN_1V, AIN_2V5, AIN_5V or AIN_10V)
 *   avg     = Number of averaging times (AVG_1, AVG_2, AVG_4, AVG_8 or AVG_16)
 *
 * Return value: Analog value (0-4095, 0 = 0V, 4095 = max range)
 */
int PV2019_AIN_Volt (int base, int ch, int range, int avg)
{
  ch    &= 0x000F;
  range &= 0x3000;
  avg   &= 0x0007;
  return PV2019_Cmd (base, (unsigned short)((ch<<8) | range | avg));
}



/*---------------------------------------------------------------------------
 * PV2019_AIN_Curr(): Read one of the analog current inputs (AIN 16-23).
 *
 * Parameters:
 *   base    = base address of the PV-2019 adapter
 *   channel = AIN input number (16-23)
 *   avg     = Number of averaging times (AVG_1, AVG_2, AVG_4, AVG_8 or AVG_16)
 *
 * Return value: Analog value (0-4095, 0 = 0mA, 4095 = 20mA)
 */
int PV2019_AIN_Curr (int base, int ch, int avg)
{
  ch  &= 0x0007;
  avg &= 0x0007;
  return PV2019_Cmd (base, (unsigned short)((ch<<8) | avg | 0x4000u));
}



/*---------------------------------------------------------------------------
 * PV2019_AOUT(): Set an analog output.
 *
 * Parameters:
 *   base  = base address of the PV-2019 adapter
 *   ch    = AOUT channel (0 or 1)
 *   range = output range (AOUT_0_10V, AOUT_0_20MA or AOUT_4_20MA)
 *   value = output value (0-4095, 0=min range, 4095=max range)
 */
void PV2019_AOUT (int base, int ch, int range, int value)
{
  ch    &= 0x0001;
  range &= 0xE000;
  value &= 0x0FFF;
  PV2019_Cmd (base, (unsigned short)((ch<<12) | range | value));
}



/*---------------------------------------------------------------------------
 * PV2019_CntEnable(): Start and read one of the two counters
 *
 * Parameters:
 *   base  = base address of the PV-2019 adapter
 *   cnt   = number of the counter (0 or 1)
 *
 * Return value: the current 32-bit value of the counter
 */
unsigned long PV2019_CntEnable (int base, int cnt)
{
  unsigned long value;

  /* Enable the counter and read the low word, and read the high word */
  cnt &= 0x0001;
  value = PV2019_Cmd (base, (unsigned short)(cnt<<11) | 0x5000u);
  return value | (PV2019_Cmd (base, (unsigned short)(cnt<<11) | 0x5300u) << 16);
}



/*---------------------------------------------------------------------------
 * PV2019_CntDisable(): Stop and read one of the two counters
 *
 * Parameters:
 *   base  = base address of the PV-2019 adapter
 *   cnt   = number of the counter (0 or 1)
 *
 * Return value: the current 32-bit value of the counter
 */
unsigned long PV2019_CntDisable (int base, int cnt)
{
  unsigned long value;

  /* Disable the counter and read the low word, and read the high word */
  cnt &= 0x0001;
  value = PV2019_Cmd (base, (unsigned short)(cnt<<11) | 0x5100u);
  return value | (PV2019_Cmd (base, (unsigned short)(cnt<<11) | 0x5300u) << 16);
}



/*---------------------------------------------------------------------------
 * PV2019_CntReset(): Reset one of the two counters
 *
 * Parameters:
 *   base  = base address of the PV-2019 adapter
 *   cnt   = number of the counter (0 or 1)
 */
void PV2019_CntReset (int base, int cnt)
{
  cnt &= 0x0001;
  PV2019_Cmd (base, (unsigned short)((cnt<<11) | 0x5200u));
}



/*---------------------------------------------------------------------------
 * PV2019_CntStatus(): Read the status of the two counter inputs as digital
 * inputs.
 *
 * Parameters:
 *   base  = base address of the PV-2019 adapter
 *
 * Return value:
 *   Bit 0 = counter 0 input status (0 or 1)
 *   Bit 1 = counter 1 input status (0 or 1)
 */
unsigned char PV2019_CntStatus (int base)
{
  return (unsigned char) PV2019_Cmd (base, 0x4E00u);
}
